/************************************************************************
*                                                                       *
* System      : CNTPIX readout module                                   *
* Block        : Example for CNTPIX                                     *
* Module      : SOIPIX_USER                                             *
* Version     : v 0.0.1 2008/10/09                                      *
*                                                                       *
* Description : Top Module                                              *
*                                                                       *
* Designer    : Minoru Hirose                                           *
*                                                                       *
*                Copyright (c) 2008 Tomohisa Uchida  & Minoru Hirose    *
*                All rights reserved                                    *
*                                                                       *
************************************************************************/
`define FPGA_VER 32'h0805_1901
`define FPGA_ID  32'h0005_0000

//`include "../Src/sys_dcm.v"
//`include "../Src/SIO_SLAVE.v"
//`include "../Src/SIO_REG.v"

module SOIPIX_USER(
// System
	OSC					,	// in	: 50MHz clock
	SW_RSTn				,	// in	: System reset
// User I/F
	USR_CLK				,	// out	: Clock
	USR_ACTIVE			,	// in	: TCP established
	USR_CLOSE_REQ		,	// in	: Request to close a TCP Connection
	USR_CLOSE_ACK		,	// out	: Acknowledge for USR_CLOSE_REQ
// TCP Tx
	USR_TX_AFULL		,	// in	: Almost full flag of a TCP Tx FIFO
	USR_TX_WE			,	// out	: TCP Tx Data write enable
	USR_TX_WD			,	// out	: TCP Tx Data[7:0]
// TCP Rx
	USR_RX_EMPTY		,	// in	: Empty flag of a TCP RX flag
	USR_RX_RE			,	// out	: TCP Rx Data Read enable
	USR_RX_RV			,	// in	: TCP Rx data valid
	USR_RX_RD			,	// in	: TCP Rx data[7:0]
// Register Access
	REG_CLK				,	// in	: Clock
	REG_ACT				,	// in	: Access Active
	REG_DI				,	// out	: Serial Data Output
	REG_DO				,	// in	: Serial Data Input
//DAC test
	DAC_D             ,  // out   : DAC DATA
	DAC_SYNC          ,  // out   : DAC SYNC
	DAC_SCLK          ,  // out   : DAC CLK
//CNTPIX2 Signals
	CSR					,
	WR						,
	RA						,
	CA						,
	RAEN					,
	CAEN					,
	ENDOUT				,
	DI						,
	DO						,
	PIXCNT_RST			,
	ENVL2_X           ,
// SW, LED
	LED					,	// out	: LED[7:0]
	DIP_SW					// in	: DIP SW[7:0]
);

//-------- Input/Output -------------
	input			 OSC				  ;
	input			 SW_RSTn			  ;

	output		   USR_CLK		  ;
	input			 USR_ACTIVE		  ;
	input			 USR_CLOSE_REQ	  ;
	output		   USR_CLOSE_ACK ;
	input			 USR_TX_AFULL	  ;
	output		   USR_TX_WE	  ;
	output [7:0]   USR_TX_WD	  ;
	input			 USR_RX_EMPTY	  ;
	output		   USR_RX_RE	  ;
	input			 USR_RX_RV		  ;
	input	[7:0]	 USR_RX_RD		  ;
	input			   REG_CLK		  ;
	input			 REG_ACT			  ;
	output		   REG_DI	 	  ;
	input			 REG_DO			  ;
	output   		DAC_D   	     ;
	output   		DAC_SYNC      ;
	output   		DAC_SCLK      ;
	//CNTPIX2 Signals
	output 			CSR			;
	output			WR				;
	output	[6:0]	RA				;
	output 	[6:0]	CA				;
	output 			RAEN			;
	output 			CAEN			;
	output 			ENDOUT		;
	output	[12:4]	DI			;
	input 	[15:0]	DO			;
	output 			PIXCNT_RST	;
   output         ENVL2_X     ;
 
	output	[7:0]	LED			;
	input	[7:0]	DIP_SW			;

//----------------------------------------------------------------
//	DLL
//----------------------------------------------------------------
	wire			RST_DLL				;
	wire			locked				;
	wire			CLK25M				;
	wire			CLK50M				;

	assign	RST_DLL		= ~SW_RSTn;

   	sys_dcm			SYSDCM(
		.CLKIN_IN			(OSC			),
		.RST_IN				(RST_DLL		),
		.CLKDV_OUT			(CLK25M			),
		.CLKIN_IBUFG_OUT	(),
		.CLK0_OUT			(CLK50M			),
		.LOCKED_OUT			(locked			)
	);

	assign	RSTn	= locked & ~RST_DLL;

	ODDR	IOB_DDRCLKp0(.Q(USR_CLK), .C(CLK25M), .CE(1'b1), .D1(1'b0), .D2(1'b1), .R( 1'b0), .S( 1'b0));

//-------------------------------------------------------------------------
//	Contents of Reg => LED
//-------------------------------------------------------------------------
	wire	[7:0]	LED			;
	reg	[7:0]	ledOut		;

	reg [7:0] regaddr;
	wire [7:0] ADDR00; //DAQ Flag 0: DAQ, 1: wait
	wire [7:0] ADDR01; //DAQ Interbal [23:16]
	wire [7:0] ADDR02; //DAQ Interbal [15:8]
	wire [7:0] ADDR03; //DAQ Interbal [7:0]
	wire [7:0] ADDR04; //RAEN Hi Time [23:16]
	wire [7:0] ADDR05; //RAEN Hi Time [15:8]
	wire [7:0] ADDR06; //RAEN Hi Time [7:0]
	wire [7:0] ADDR07; //CSR initial value (except TIN)
	wire [7:0] ADDR08; //DAC_control [15:8]
	wire [7:0] ADDR09; //DAC_control [7:0]
	wire [7:0] ADDR10; //RA select
	wire [7:0] ADDR11; //CA select
	reg	[7:0] ADDR [0:11]		;	

	always@ (posedge CLK25M)begin
		if(~RSTn)begin
			regaddr <= 8'b0;
		end
		regaddr <= DIP_SW;
		ADDR[0] <= ADDR00;
		ADDR[1] <= ADDR01;
		ADDR[2] <= ADDR02;
		ADDR[3] <= ADDR03;
		ADDR[4] <= ADDR04;
		ADDR[5] <= ADDR05;
		ADDR[6] <= ADDR06;
		ADDR[7] <= ADDR07;
		ADDR[8] <= ADDR08;
		ADDR[9] <= ADDR09;
		ADDR[10] <= ADDR10;
		ADDR[11] <= ADDR11;
		case(regaddr)
			8'h0:	ledOut <= ADDR[0];
			8'h1:	ledOut <= ADDR[1];
			8'h2:	ledOut <= ADDR[2];
			8'h3:	ledOut <= ADDR[3];
			8'h4:	ledOut <= ADDR[4];
			8'h5:	ledOut <= ADDR[5];
			8'h6:	ledOut <= ADDR[6];
			8'h7:	ledOut <= ADDR[7];
			8'h8:	ledOut <= ADDR[8];
			8'h9:	ledOut <= ADDR[9];
			8'ha: ledOut <= ADDR[10];
			8'hb: ledOut <= ADDR[11];
			default:	ledOut <= 8'b0;
		endcase
	end		
	assign LED[6:0] = ledOut[6:0];
 	reg tmpled;
 	assign LED[7] = tmpled;

	wire [7:0] wireADDR7;
	assign wireADDR7 = ADDR[7];

//DAC test-------------------------------------
	wire DAC_D;
	wire DAC_SYNC;
	wire DAC_SCLK;

	reg [20:0] dac_cnt;
	reg [15:0] dac_buf;
	reg        dac_read;

// 13bit on = 4096
	always @(negedge CLK25M)begin
//      dac_cnt <=21'b0;
//      dac_buf <=16'b0;
		if(~RSTn)begin
			dac_cnt  <= 21'b0;
			dac_buf  <= 16'b0001_0000_0000_0000;
			dac_read <= 1'b0;
		end 
		if(dac_cnt <= 21'h11)begin
			dac_read <= 1'b0;
			if(dac_cnt >= 21'h1)begin
				dac_buf <= dac_buf << 1;
			end
		end
		if(dac_cnt == 21'h1F_FFFF)begin
			dac_read <= 1'b1;
//			dac_buf <= 16'b0001_0111_1111_1111;
			dac_buf <= {ADDR[8], ADDR[9]};
		end
		dac_cnt <= dac_cnt + 21'b1;				
	end //always
	
	assign DAC_SCLK = ~CLK25M;
	assign DAC_D = dac_buf[15];
	assign DAC_SYNC = dac_read;

//-----------------------------------------------------------
//-----------------------------------------------------------
	reg				irUsrActive		;
	reg				irUsrCloseReq	;
	reg				irUsrTxAfull	;
	reg				irUsrRxEmpty	;
	reg				irUsrRxRv		;
	reg		[7:0]	irUsrRxRd		;

	always@ (posedge CLK25M) begin
		irUsrActive		<= USR_ACTIVE;
		irUsrCloseReq	<= USR_CLOSE_REQ;
		irUsrTxAfull	<= USR_TX_AFULL;
		irUsrRxEmpty	<= USR_RX_EMPTY;
		irUsrRxRv		<= USR_RX_RV;
		irUsrRxRd[7:0]	<= USR_RX_RD[7:0];
	end

//CNTPIX2 test-----------------------------------------------
	reg orCSR;
	reg orWR;
	reg orRAEN;
	reg orCAEN;
	reg [6:0] orRA;
	reg [6:0] orCA;
	reg orENDOUT;
	reg [12:4] orDI;

	reg write_flag;
	reg write_flag_first;
	reg read_flag;
	reg [2:0] write_cycle_cnt;
	reg [4:0] read_cycle_cnt;
	integer pix_size;

	reg TCP_DATA_OK;

	always @(posedge CLK25M)begin

// not active, reset parameters. 
// first, write_flag ON / read_flag OFF

		if(~irUsrActive)begin
			orCSR <= 1'b0;
			orWR <= 1'b0;
			orRAEN <= 1'b0;
			orCAEN <= 1'b0;
			orRA <= 7'b0;
			orCA <= 7'b0;
	 		orENDOUT <= 1'b0;
         orDI[12:4] <=9'b0_0000_0000;
			write_flag <= 1'b1;
			read_flag <= 1'b0;
			write_cycle_cnt <= 3'b0;	
			read_cycle_cnt <= 5'b0;	
			TCP_DATA_OK <= 1'b0;
			pix_size <= 7'd127;

// UseActive and...

		end else if(write_flag)begin
	 		orDI[12:4] <= {wireADDR7[7], 1'b0, wireADDR7[6:0]};
			orCSR <= 1'b1;
			if(write_cycle_cnt == 0) begin
			  orRA <= ADDR[10];
			  orCA <= ADDR[11];
			end
			write_cycle_cnt <= write_cycle_cnt + 3'b1;		
			case(write_cycle_cnt)
				3'h1 : begin
					orRAEN <= 1'b1;
					orCAEN <= 1'b1;
				end
				3'h2 : begin
					orWR <= 1'b1;
				end
				3'h4 : begin
					orWR <= 1'b0;
				end
				3'h5 : begin
					orRAEN <= 1'b0;
					orCAEN <= 1'b0;
				end
				3'h6 : begin
// pix cycle...				
               write_cycle_cnt<=3'b0;
//					if(orRA == pix_size & orCA == pix_size) begin
						//orCSR <= 1'b0;
						orWR <= 1'b0;
						write_flag <= 1'b0;
						read_flag <= 1'b1;
//						orRA <= 7'b0;
//						orCA <= 7'b0;
//						orDI[12:4] <= 9'b0;
//					end else if(orCA == pix_size)begin
//						orCA <= 7'b0;
//						orRA <= orRA + 7'b1;
//					end else begin
//						orCA <= orCA + 7'b1;
//					end
					
				end // 3'h6
			endcase
// active and...			
		end else if(read_flag)begin
			read_cycle_cnt <= read_cycle_cnt + 5'b1;		
			orCSR <= 1'b1;			
			case(read_cycle_cnt)
				5'h15 : begin
					orRAEN <= 1'b1;
					orCAEN <= 1'b1;
				end
				5'h18 : begin
					TCP_DATA_OK <= 1'b1;
				end
				5'h19 : begin
					TCP_DATA_OK <= 1'b0;
				end
				5'h1b : begin
					orRAEN <= 1'b0;
					orCAEN <= 1'b0;
				end
				5'h1f : begin
				   read_cycle_cnt <= 5'b0;	
					if(orRA == pix_size & orCA == pix_size) begin
						write_flag <= 1'b0;
						read_flag <= 1'b0;
						orRA <= 7'b0;
						orCA <= 7'b0;
						orCSR <= 1'b0;
						TCP_DATA_OK <= 1'b0;
//					end else if(orCA == pix_size)begin
//						orCA <= 7'b0;
//						orRA <= orRA + 7'b1;
					end else begin
//						orCA <= orCA + 7'b1;
                  read_flag  <= 1'b0;
                  write_flag <= 1'b1;
					end
				end
			endcase
		end else begin // usractive and ~read and ~write flag
//       write_flag <= 1'b1;
//			read_flag <=1'b0;
      end
	end // clk

//-------------------------------
// no write&read flag...

	reg DAQ_flag;	

	always @(posedge CLK25M)begin
		if(~irUsrActive)begin
			DAQ_flag <= 1'b0;
		end
		if(ADDR[0] == 8'b1 & ~(write_flag | read_flag))begin
			DAQ_flag <= 1'b1;
		end else begin
			DAQ_flag <= 1'b0;
		end
		tmpled <= 1'b0;
	end	

//-------------------------------

	reg [6:0] DAQ_orRA;
	reg [6:0] DAQ_orCA;
	reg DAQ_orRAEN;
	reg DAQ_orCAEN;
	reg orPIXCNT_RST;
	reg orENVL2_X;

	reg [31:0] DAQ_cycle;
	reg internal_trig;

	reg [31:0] DAQ_INTERBAL;
	integer PIX_CYCLE;
	reg send_start;
	reg DO_ready;
	reg [5:0] send_No;
	reg [14:0] send_cnt;
	reg [15:0] orDO;
   reg [31:0] CNT_TIME;

	always @(posedge CLK25M)begin
		if(~irUsrActive)begin
			internal_trig <= 1'b0;
			orPIXCNT_RST <= 1'b0;	
         CNT_TIME <= {ADDR[4], ADDR[5], ADDR[6], 4'b0};
			DAQ_INTERBAL <= CNT_TIME + {ADDR[1],ADDR[2],ADDR[3]};
			DAQ_cycle <= 32'b0;
		end 
//---
		if(DAQ_flag & ~send_start)begin
			internal_trig <= 1'b0;
			DAQ_cycle <= DAQ_cycle + 32'b1;
         CNT_TIME <= {ADDR[4], ADDR[5], ADDR[6],4'b0};
			DAQ_INTERBAL <= CNT_TIME + {ADDR[1],ADDR[2],ADDR[3]};
			
			if(DAQ_cycle == DAQ_INTERBAL-CNT_TIME)begin
				orPIXCNT_RST <= 1'b1;
			end else begin
				orPIXCNT_RST <= 1'b0;
			end
			if(DAQ_cycle == DAQ_INTERBAL)begin
				DAQ_cycle <= 32'b0;
				internal_trig <= 1'b1;
			end
//--- ~DAQ_flag or send_start
		end else begin
			if(~DAQ_flag) begin
			  internal_trig <= 1'b0;
			end
			orPIXCNT_RST <= 1'b0;
         CNT_TIME <= {ADDR[4], ADDR[5], ADDR[6],4'b0};
			DAQ_INTERBAL <= CNT_TIME + {ADDR[1],ADDR[2],ADDR[3]};
			DAQ_cycle <= 32'b0;			
		end
	end // 

	always @(posedge CLK25M)begin
		if(~irUsrActive)begin
			DAQ_orRA <= 7'b0;
			DAQ_orCA <= 7'b0;
			DAQ_orRAEN <= 1'b0;
			DAQ_orCAEN <= 1'b0;
			send_start <= 1'b0;
			DO_ready <= 1'b0;
			send_No <= 5'b0;
			send_cnt <= 15'b0;
			orDO <= 16'b0;
			// pix_cycle >=5
			PIX_CYCLE <= 5'h6;
			orENVL2_X <= 1'b0;
		end
		if(internal_trig)begin
			send_start <= 1'b1;
		end
		if(send_start & ~irUsrTxAfull)begin
		   if(send_No == 0 & DAQ_orCA == 0 & DAQ_orRA == 0) begin
			   orENVL2_X <= 1'b1;
			end
			if(send_No == 0 ) begin
			  DO_ready <= 1'b0;
			end
		
		   send_No <= send_No + 5'b1;
			case(send_No)
				5'h0 : begin
					DO_ready <= 1'b0;
				end
				5'h1: begin
					DAQ_orRAEN <= 1'b1;
					DAQ_orCAEN <= 1'b1;			
				end				
				// 3-->4 --> 3
				PIX_CYCLE-5'd3 : begin
					orDO <= DO;
				end
				PIX_CYCLE-5'd2 : begin
					DAQ_orCAEN <= 1'b0;
					if(DAQ_orCA == pix_size)begin
						DAQ_orRAEN <= 1'b0;
					end
					DO_ready <= 1'b1;
				end
				PIX_CYCLE-5'd1 : begin
					DO_ready <= 1'b1;
				end
				PIX_CYCLE : begin
					send_cnt <= send_cnt + 15'b1;
					DO_ready <= 1'b0;
					if(DAQ_orCA == pix_size)begin
						DAQ_orCA <= 7'b0;
						DAQ_orRA <= DAQ_orRA + 7'b1;
					end else begin
						DAQ_orCA <= DAQ_orCA + 7'b1;
					end
					send_No <= 5'b0;
				end
			endcase
			if(send_cnt == 15'b100_0000_0000_0000-15'b1 & send_No == PIX_CYCLE)begin
				send_cnt <= 15'b0;
				send_No <= 5'b0;
				send_start <= 1'b0;
				DAQ_orRA <= 7'b0;
				DAQ_orCA <= 7'b0;
				DAQ_orRAEN <= 1'b0;
				DAQ_orCAEN <= 1'b0;
				DO_ready <= 1'b0;
				orDO <= 16'b0;
            orENVL2_X <= 1'b0;				
			end
// ~sent_start or AXFULL
		end else begin
			DO_ready <= 1'b0;
		end			
	end // clk

	assign CSR = orCSR;
	assign WR = orWR;
	assign RAEN = orRAEN | DAQ_orRAEN;
	assign CAEN = orCAEN | DAQ_orCAEN;
	assign RA = orRA | DAQ_orRA;
	assign CA = orCA | DAQ_orCA;
	assign DI[12:4] = orDI[12:4];
	assign ENDOUT = read_flag | DAQ_flag | send_start;
	assign PIXCNT_RST = orPIXCNT_RST;
   assign ENVL2_X = orENVL2_X;

//--------------------------------------------------------------------------
//TCP section---------------------------------------------------------------
	reg				usrTxWe		;
	reg		[7:0]	usrTxWd		;
	reg				usrRxRe		;

	always@ (posedge CLK25M) begin
		usrTxWe		<=  irUsrActive & ~irUsrCloseReq & (TCP_DATA_OK | DO_ready);
//---csr transfer
		if(read_flag & read_cycle_cnt == 5'h19)begin
			//usrTxWd[7:0]	<=  {DO[12:8], DO[6:4]};
			usrTxWd[7:0]	<=  DO[11:4];
//---data transfer
		end else if(send_start)begin
			if(send_No == PIX_CYCLE-5'h1)begin
				//usrTxWd[7:0]	<=  orDO[15:8];
				usrTxWd[7:0]	<=  orDO[7:0];
			   //usrTxWd[7:0]   <=  8'b0;	
			end else if(send_No == PIX_CYCLE)begin
				usrTxWd[7:0]	<=  orDO[15:8];
				//usrTxWd[7:0]	<=  orDO[7:0];
				//usrTxWd[7:0]   <= 8'b0;
			end else begin
				usrTxWd[7:0] <= 8'b0;
			end
//--- not csr nor data
		end else begin
      		usrTxWd[7:0] <= 8'b0;
		//	usrTxWd[7:0]	<=  8'b10101010;
		end			
		usrRxRe		<=  irUsrActive & ~irUsrCloseReq & ~irUsrRxEmpty;
	end

	reg				orUsrCloseAck	;
	reg				orUsrTxWe		;
	reg		[7:0]	orUsrTxWd		;
	reg				orUsrRxRe		;

	always@ (posedge CLK25M) begin
		orUsrCloseAck	<= irUsrCloseReq;
		orUsrTxWe		<= usrTxWe;
		orUsrTxWd[7:0]	<= usrTxWd[7:0];
		orUsrRxRe		<= usrRxRe;
	end

	assign	USR_CLOSE_ACK	= orUsrCloseAck;
	assign	USR_TX_WE		= orUsrTxWe;
	assign	USR_TX_WD		= orUsrTxWd[7:0];
//	assign	USR_TX_WD		= 8'b0;
	assign	USR_RX_RE		= orUsrRxRe;

//------------------------------------------------------------------------------
//	Register Access
//------------------------------------------------------------------------------
	wire	[31:0]	SIO_ADDR		;
	wire	[7:0]	SIO_WD			;
	wire			SIO_WE			;
	wire			SIO_RE			;
	wire			SIO_ACK			;
	wire			SIO_RV			;
	wire	[7:0]	SIO_RD			;

	SIO_SLAVE			SIO_SLAVE(
		.RSTn					(RSTn					),	// in	: System reset
		.FILL_ADDR			(32'h0					),	// in	: Filled address for narow address-width
		// Serial I/F
		.SCK					(REG_CLK				),	// in	: Clock
		.SCS					(REG_ACT				),	// in	: Active
		.SI					(REG_DI					),	// out	: Data input
		.SO					(REG_DO					),	// in	: Data output
		// Register I/F
		.REG_ADDR			(SIO_ADDR[31:0]			),	// out	: Address[31:0]
		.REG_WD				(SIO_WD[7:0]			),	// out	: Data[7:0]
		.REG_WE				(SIO_WE					),	// out	: Write enable
		.REG_RE				(SIO_RE					),	// out	: Read enable
		.REG_ACK				(SIO_ACK				),	// in	: Access acknowledge
		.REG_RV				(SIO_RV					),	// in	: Read valid
		.REG_RD				(SIO_RD[7:0]			)	// in	: Read data[7:0]
	);

	SIO_REG				SIO_REG(
		.SCK					(REG_CLK				),	// in	: Clock
		.RSTn					(RSTn					),	// in	: System reset
		//test by hirose
		.ADDR00						(ADDR00				),
		.ADDR01						(ADDR01				),
		.ADDR02						(ADDR02				),
		.ADDR03						(ADDR03				),
		.ADDR04						(ADDR04				),
		.ADDR05						(ADDR05				),
		.ADDR06						(ADDR06				),
		.ADDR07						(ADDR07				),
		.ADDR08						(ADDR08				),
		.ADDR09						(ADDR09				),
		.ADDR10                 (ADDR10           ),
		.ADDR11                 (ADDR11           ),
		// Register I/F	
		.REG_ADDR				(SIO_ADDR[31:0]			),	// in	: Address[31:0]
		.REG_WD					(SIO_WD[7:0]			),	// in	: Data[7:0]
		.REG_WE					(SIO_WE					),	// in	: Write enable
		.REG_RE					(SIO_RE					),	// in	: Read enable
		.REG_ACK				(SIO_ACK				),	// out	: Access acknowledge
		.REG_RV					(SIO_RV					),	// out	: Read valid
		.REG_RD					(SIO_RD[7:0]			)	// out	: Read data[7:0]
	);

//------------------------------------------------------------------------------
endmodule